import logging
import json

from collections import namedtuple
from typing import List, Dict, Optional, Union, Any

import requests
from requests.compat import urljoin  # type: ignore
from requests.utils import prepend_scheme_if_needed

from ppa import appliance
from ppa_appliance import vaults

from . import exceptions

logger = logging.getLogger(__name__)


def _assert_engine_version(engine_version):
    if engine_version not in {1, 2}:
        raise exceptions.InvalidEngineVersion(f"The kv_version must be 1 or 2. Received {engine_version}.")


def proxies():
    return {"https": appliance.api_proxy(), "http": appliance.api_proxy()}


def verify():
    return appliance.cert()


def _get_vault(reason: str, vault_name: Optional[str] = None) -> vaults.Vault:
    return (
        vaults.hashicorp_vault_by_name(vault_name, reason=reason)
        if vault_name
        else vaults.select_vault(vaults.hashicorp_vaults(reason=reason))
    )


def _form_secret_url(vault_url: str, secret: str, engine: str, kv_version: int) -> str:
    return urljoin(
        prepend_scheme_if_needed(vault_url, "https"),
        f"/v1/{engine}/{secret}" if kv_version == 1 else f"/v1/{engine}/data/{secret}",
    )


def _parse_response(response: requests.Response) -> dict:
    try:
        return response.json()
    except json.decoder.JSONDecodeError:
        message = f"could not convert the following response body to JSON: {response.text}"
        logger.error(message)
        raise exceptions.VaultError(message)


def _get_errors(response_json: Any) -> Union[str, None]:
    combined_errors = ", ".join((response_json.get("warnings") or []) + (response_json.get("errors") or []))
    if not combined_errors:
        return None
    return combined_errors


def _unpack_and_raise(
    action: str, secret: str, engine: str, kv_version: int, response_json: Any, status_code: int
) -> None:
    logger.debug(f"{response_json=}")
    combined_errors = _get_errors(response_json)
    logger.error(f"status code: {status_code}, combined errors: {combined_errors}")

    if status_code == 404:
        if not combined_errors:
            raise exceptions.InvalidSecret(
                f"Failed to {action} secret data. Supplied secret '{secret}' does not "
                f"exist in v{kv_version} engine '{engine}'."
            )
        if "no handler for route" in combined_errors:
            raise exceptions.InvalidEngine(
                f"Failed to {action} secret data. Supplied v{kv_version} engine '{engine}' does not exist."
            )
        elif "Invalid path for a versioned K/V secrets engine" in combined_errors:
            raise exceptions.InvalidSecret(
                f"Failed to {action} secret data. Supplied secret '{secret}' exists but is held in a KV "
                f"engine of a different version. The operation was attempted using the v{kv_version} API."
            )

    if status_code == 503:
        if combined_errors and "Vault is sealed" in combined_errors:
            raise exceptions.VaultSealed(f"Failed to {action} secret data. The Hashicorp Vault is sealed.")

    raise exceptions.VaultError(
        f"Failed to {action} keys against Hashicorp Vault secret '{secret}' using v{kv_version} engine "
        f"'{engine}'. Status code {status_code}. Please view logs for more information."
    )


def _filter_keys(data: Dict[str, str], keys: List[str], secret: str) -> Dict[str, str]:
    """
    - Takes a secret dictionary and a list of keys.
    - Returns a dictionary that only contains the requested keys.
    """
    missing_keys = [key for key in keys if key not in data]
    if missing_keys:
        raise exceptions.InvalidKey(
            "Failed to read secret data. Requested key(s) '{0}' not found in secret '{1}'.".format(
                ", ".join(missing_keys), secret
            )
        )

    return {key: value for key, value in data.items() if key in keys}


def _load_secret_record(data: dict):
    """ We don't know which fields will exist in each secret until either we list the secret, or the user
    supplies the keys. For that reason we have to create a namedtuple based class on the fly.
    """
    fields = list(data.keys())
    logger.debug(f"creating VaultSecret record with fields {fields}")

    # Silence the following mypy error:
    # 'error: List or tuple literal expected as the second argument to namedtuple()'
    # There's no way I'm defining a list twice just to make mypy happy.
    VaultSecret = namedtuple("HashicorpVaultSecret", fields)  # type: ignore

    # Silence the following mypy error 'error: Too many arguments for "VaultSecret"'
    return VaultSecret(*[data.get(field) for field in fields])  # type: ignore


def _create_engine(
    *,
    engine: str,
    kv_version: Optional[int] = 2,
    reason: Optional[str] = "Task is creating a new secrets engine in Hashicorp Vault",
    vault_name: Optional[str] = None,
):
    logger.info(f"attempting to create v{kv_version} secrets engine '{engine}'")
    hashicorp_vault = _get_vault(reason, vault_name=vault_name)
    response = requests.post(
        urljoin(prepend_scheme_if_needed(hashicorp_vault.url, "https"), f"/v1/sys/mounts/{engine}"),
        proxies=proxies(),
        verify=verify(),
        json={"type": "kv", "options": {"version": str(kv_version)}},
    )
    if response.status_code != 204:
        errors = _get_errors(_parse_response(response))
        if "already in use" in errors:
            raise exceptions.EngineAlreadyExists(f"A secrets engine already exists with supplied path '{engine}'.")
        logger.error(response.status_code)
        logger.error(response.json())
        raise exceptions.VaultError(
            f"Failed to create v{kv_version} secrets engine in path '{engine}'. Status code "
            f"{response.status_code}. Please view the logs for more information."
        )


def _read_secret(
    *,
    secret: str,
    engine: str,
    keys: Optional[List[str]] = None,
    reason: Optional[str] = "Task is accessing keys from Hashicorp Vault",
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
):
    """ List a Hashicorp Vault secret and return only the values for the supplied keys.
    """
    _assert_engine_version(kv_version)

    logger.info(f"attempting to read keys from secret {engine}/{secret}")
    hashicorp_vault = _get_vault(reason, vault_name=vault_name)

    response = requests.get(
        _form_secret_url(hashicorp_vault.url, secret, engine, kv_version), proxies=proxies(), verify=verify()
    )

    response_json = _parse_response(response)

    if response.status_code != 200:
        _unpack_and_raise("read", secret, engine, kv_version, response_json, response.status_code)

    try:
        records = response_json["data"] if kv_version == 1 else response_json["data"]["data"]
        assert records
    except (KeyError, AssertionError):
        raise exceptions.EmptySecret(
            f"Hashicorp Vault secret '{secret}' in engine '{engine}' exists but contains no data."
        )
    return _load_secret_record(_filter_keys(records, keys, secret) if keys else records)


def _create_secret(
    *,
    secret: str,
    engine: str,
    data: dict,
    reason: Optional[str] = "Task is creating a new secret in Hashicorp Vault",
    vault_name: Optional[str] = None,
    overwrite: Optional[bool] = True,
    kv_version: Optional[int] = 2,
) -> None:
    """ Create a new secret in a Hashicorp Vault KV engine.

    This is very similar to update, but does not try to list & combine existing keys in the secret first.

    If the secret already exists, the operation will fail unless overwrite is set to True.

    When overwrite is set to True, all existing keys in the secret will be replaced or deleted.
    """
    _assert_engine_version(kv_version)

    try:
        existing_secrets = _read_secret(
            secret=secret,
            engine=engine,
            reason="Task is checking whether secret exists before creation",
            vault_name=vault_name,
            kv_version=kv_version,
        )
        logger.debug(f"{existing_secrets=}")
        if overwrite is False:
            raise exceptions.SecretAlreadyExists(
                f"Secret '{secret}' already exists in v{kv_version} engine '{engine}'."
            )
        logger.info("overwrite is true, so the existing secret will be replaced with the new data")
    except exceptions.InvalidSecret:
        logger.info("secret does not exist and will be created")

    hashicorp_vault = _get_vault(reason, vault_name=vault_name)
    secret_url = _form_secret_url(hashicorp_vault.url, secret, engine, kv_version)

    response = requests.post(
        secret_url, proxies=proxies(), verify=verify(), json=data if kv_version == 1 else {"data": data}
    )

    if any([all([kv_version == 1, response.status_code != 204]), all([kv_version == 2, response.status_code != 200])]):
        _unpack_and_raise("create", secret, engine, kv_version, _parse_response(response), response.status_code)


def _update_secret(
    *,
    secret: str,
    engine: str,
    data: Dict[str, str],
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
) -> None:
    """ Update one or more keys in a Hashicorp Vault secret.

    If the secret doesn't exist, this function will raise.

    Create must be used to create new secrets. This allows delegation of responsibility.
    """
    _assert_engine_version(kv_version)

    def _get_existing_secrets():
        try:
            existing_secrets = _read_secret(
                secret=secret,
                engine=engine,
                reason="Task is getting existing secret values before update",
                kv_version=kv_version,
                vault_name=vault_name,
            )
        except exceptions.InvalidSecret:
            raise exceptions.InvalidSecret(
                f"Secret '{secret}' was not found in v{kv_version} engine '{engine}'. "
                "Update failed. Consider using create_secret instead."
            )
        except exceptions.EmptySecret:
            # The secret exists but is empty, make an empty vault secrets record.
            existing_secrets = _load_secret_record({})
        logger.debug(f"secret has {len(existing_secrets)} existing keys that will be copied to the new secret")
        return existing_secrets

    def _build_update_data(existing_secrets, new_records: Dict[str, str]) -> dict:
        updated_secret = existing_secrets._asdict()

        for key, value in new_records.items():
            logger.debug(f"adding supplied value for {key} into secret payload")
            updated_secret[key] = value

        return updated_secret if kv_version == 1 else {"data": updated_secret}

    update_count = len(data)
    hashicorp_vault = _get_vault(
        reason or f"Task is updating {update_count} keys in Hashicorp Vault", vault_name=vault_name
    )
    secret_url = _form_secret_url(hashicorp_vault.url, secret, engine, kv_version)
    logger.info(f"attempting to update {update_count} keys in v{kv_version} engine using path {secret_url}")

    updated_secret = _build_update_data(_get_existing_secrets(), data)

    response = requests.post(secret_url, proxies=proxies(), verify=verify(), json=updated_secret)

    if any([all([kv_version == 1, response.status_code != 204]), all([kv_version == 2, response.status_code != 200])]):
        _unpack_and_raise("update", secret, engine, kv_version, _parse_response(response), response.status_code)
